﻿using RestSharp;
using System;
using System.Threading.Tasks;
using Newtonsoft;
using Newtonsoft.Json;
using System.Linq;
using System.Collections.Generic;

namespace Optiroute.DataLoader.Google
{
    /// <summary>
    /// Obsahuje task který vrací základní info o adrese v době kdy doběhne metoda RecognizeAddess
    /// </summary>
    public interface IAddressRecognizer
    {
        Task<AddressResult> RecognizeAddress(string queryAddress);
        Task<RouteMatrixResult> FindRouteLenght(FindRouteModel findRouteModel);
    }

    /// <summary>
    /// Servisa, která umožňuje na základě textu adresy vrátit FullAddress a PlaceID
    /// </summary>
    public class AddressGoogleServices : IAddressRecognizer
    {
        private readonly IRestClient _client;

        /// <summary>
        /// Konstruktor
        /// </summary>
        /// <param name="apiKey">apiKey je klíč, získaný při regisraci do google geocode API, je potřeba ho mít jako parametr v kažném requestu na geocode api</param>
        /// <param name="baseURL">baseURL je první část url, která se s žádným requesem nemění. Není zde přesně definována servisa ani parametry.</param>
        public AddressGoogleServices(string apiKey, string baseURL)
        {
            if (string.IsNullOrEmpty(apiKey))
            {
                throw new ArgumentException($"'{nameof(apiKey)}' cannot be null or empty.", nameof(apiKey));
            }

            if (string.IsNullOrEmpty(baseURL))
            {
                throw new ArgumentException($"'{nameof(baseURL)}' cannot be null or empty.", nameof(baseURL));
            }

            _client = new RestClient(baseURL);
            _client.AddDefaultQueryParameter("key", apiKey);
        }

        /// <summary>
        /// Metoda umožnuje na základě textu rozeznat aresu a vrátit její celé znění a jedinečný identifikátorr adresy
        /// </summary>
        /// <param name="queryAddress">Text adresy, můž se jednat třeba jen o část nějaké adresy. Na základě totoho textu se hledají další info k adrese</param>
        /// <returns>Vrací se objekt, který obsahuje text celé adresy a ID adresy</returns>
        public async Task<AddressResult> RecognizeAddress(string queryAddress)
        {
            IRestRequest request = new RestRequest("/geocode/json", Method.GET);
            request.AddParameter("address", queryAddress, ParameterType.QueryString);

            IRestResponse response = await _client.ExecuteAsync(request);

            //Pokud bude v requestu např. queryAddress, kterou to nerozezná tak bude status kód ok (Je třeba ošetřit i jinak)
            if (response.StatusCode != System.Net.HttpStatusCode.OK)
            {
                throw new ApplicationException($"Chyba při volání Google.API geocode. Status kód: {response.StatusCode}");
            }

            GeocodingnResults objectResponse = JsonConvert.DeserializeObject<GeocodingnResults>(response.Content);

            return new AddressResult()
            {
                FullAdress = objectResponse.Results.First().FullAddress,
                PlaceID = objectResponse.Results.First().PlaceId
            };
        }

        public async Task<RouteMatrixResult> FindRouteLenght(FindRouteModel addressesToVisit)
        {
            List<List<String>> destinationList = new List<List<string>>();
            int maxItemsToCall = 25;
            for (int i = 0; i < addressesToVisit.Destinations.Count; i = i + maxItemsToCall)
            {
                int skip = i;
                int take = maxItemsToCall;
                if (i + maxItemsToCall > addressesToVisit.Destinations.Count) take = addressesToVisit.Destinations.Count - i;
                destinationList.Add(addressesToVisit.Destinations.Skip(skip).Take(take).ToList());
            }
            RouteMatrixResult routeResult = new RouteMatrixResult();

            foreach (var destinations in destinationList)
            {
                IRestRequest request = new RestRequest("/distancematrix/json", Method.GET);
                request.AddParameter("origins", string.Join('|', addressesToVisit.Origins), ParameterType.QueryString);
                request.AddParameter("destinations", string.Join('|', destinations), ParameterType.QueryString);
                request.AddParameter("mode", addressesToVisit.TravelMode.ToString().ToLower(), ParameterType.QueryString);

                IRestResponse response = await _client.ExecuteAsync(request);
                if (response.StatusCode != System.Net.HttpStatusCode.OK)
                {
                    throw new ApplicationException($"Chyba při volání Google.API distancematrix. Status kód: {response.StatusCode}");
                }
                DistancematrixResults objectResponse = JsonConvert.DeserializeObject<DistancematrixResults>(response.Content);
             
                for (int originIndex = 0; originIndex < addressesToVisit.Origins.Count(); originIndex++)
                {
                    for (int destinationIndex = 0; destinationIndex < destinations.Count(); destinationIndex++)
                    {
                        routeResult.Routes.Add(new Route()
                        {
                            Origin = new Place()
                            {
                                OriginalAddress = addressesToVisit.Origins[originIndex],
                                RecognizedAddress = objectResponse.OriginAddresses[originIndex]
                            },
                            Destination = new Place()
                            {
                                OriginalAddress = destinations[destinationIndex],
                                RecognizedAddress = objectResponse.DestinationAddresses[destinationIndex]
                            },
                            DistanceMeters = objectResponse.Rows[originIndex].Elements[destinationIndex].Distance?.Value ?? -1,
                            DurationTime = objectResponse.Rows[originIndex].Elements[destinationIndex].Duration?.Value ?? -1
                        });
                    }
                }
            }
            
            return routeResult;
        }
    }

    public class AddressResult
    {
        public string FullAdress { get; set; } //Celé znění adresy, pro vypsní uživateli
        public string PlaceID { get; set; } //ID adresy vhodné např. pro hledání tras mezi dvěma body pomocí google API
    }
    public class FindRouteModel
    {
        public FindRouteModel(List<string> origins, List<string> destinations, TravelMode travelMode)
        {
            Origins = origins;
            Destinations = destinations;
            TravelMode = travelMode;
        }

        public List<string> Origins { get; }
        public List<string> Destinations { get; }
        public TravelMode TravelMode { get; }
    }
    public enum TravelMode
    {
        Driving,
        Bicycling,
        Transit,
        Walking
    }
    /// <summary>
    /// Třída pro vrácení hodnot matice délek cest mezi adresami. Identifikace je id adresy z a do, a obsahuje délku trasy v čase i v metrech. 
    /// </summary>
    public class RouteMatrixResult
    {
        public List<Route> Routes { get; set; } = new List<Route>();
    }

    public class Route
    {
        public Place Origin { get; set; }
        public Place Destination { get; set; }
        public int DistanceMeters { get; set; }
        public int DurationTime { get; set; }
    }

    public class Place
    {
        public string OriginalAddress { get; set; }
        public string RecognizedAddress { get; set; }
    }


}
