﻿using Microsoft.AspNetCore.Mvc;
using Microsoft.Extensions.Logging;
using Newtonsoft.Json;
using Optiroute.Db;
using Optiroute.Db.Model;
using Optiroute.Web.Models;
using Optiroute.Web.Models.Home;
using Optiroute.Web.Services;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Threading.Tasks;

namespace Optiroute.Web.Controllers
{
    public class HomeController : Controller
    {
        private readonly ILogger<HomeController> _logger;
        private readonly DbContext _dbContext;
        private readonly IComputeAlgorithmRoutesService _computeService;
        private readonly ICodeBookService _codebookservice;

        public HomeController(ILogger<HomeController> logger, DbContext dbContext, IComputeAlgorithmRoutesService computeService, ICodeBookService codebookservice)
        {
            _logger = logger;
            this._dbContext = dbContext;
            this._computeService = computeService;
            this._codebookservice = codebookservice;
        }

        [HttpGet]
        public IActionResult Index()
        {
            var a = new IndexViewModel()
            {
                Addresses = _dbContext.Places.Select(d => new AddressViewModel()
                {
                    Id = d.Id,
                    NickName = d.NickName,
                    Default = d.Default,
                    EndCity = d.EndCity
                }).ToList(),

                TravelModes = _codebookservice.GetTravelModes().Select(d => new TravelModeViewModel() 
                {
                     Value = d.Travel.ToString(),
                     TravelModeText = d.TravelText,
                     Icon = d.TravelIcon
                }).ToList(),

                RouteWeights = _codebookservice.GetWeightModes().Select(d => new RouteWeightViewModel()
                {
                     Value = d.Mode.ToString(),
                     Icon = d.WeightIcon,
                     RouteWeightText =d.WeightText
                }).ToList()
            };

            a.RouteWeightValue = WeightMode.Duration.ToString();
            a.TravelModeValue = Db.Model.TravelMode.Transit.ToString();

            List<int> oftenUsedCitiesIDs = new List<int>() { 1 , 2, 3, 6, 7, 8, 9, 11, 14, 15, 16, 18, 20};
            foreach (var id in oftenUsedCitiesIDs)
            {
                a.Addresses.SingleOrDefault(d => d.Id == id).Selected = true;
            }


            return View(a);
        }

        [HttpPost]
        public IActionResult Index(IndexViewModel model)
        {
            TempData["IndexViewModel"] = JsonConvert.SerializeObject(model);
            return RedirectToAction("ResultRoutes");
        }

        public IActionResult AddressesList()
        {
            var a = new AddressesListViewModel()
            {
                Addresses = _dbContext.Places.Select(d => new AddressListViewModel()
                {
                    Id = d.Id,
                    NickName = d.NickName,
                    FullName = d.FullName,
                    FullAddress = d.FullAddress.Replace(", Czechia", ""),
                    Default = d.Default,
                    EndCity = d.EndCity
                }).ToList()
            };
            return View(a);
        }

        public IActionResult RoutesList()
        {
            var routes = new RouteListViewModel()
            {
                Routes = _codebookservice.GetTravelModes().Select(d => new RouteListSetViewModel() {
                    Icon = d.TravelIcon,
                    TravelModeText = d.TravelText,
                    Routes = GetRoutes(d.Travel) 
                }).ToList()
            };

            return View(routes);
        }

        public IActionResult ResultRoutes()
        {
            var tempData = TempData["IndexViewModel"] as string;
            if (tempData == null) return RedirectToAction("index");
            IndexViewModel model = JsonConvert.DeserializeObject<IndexViewModel>(tempData);
            var resultAlgoritmRoutes = _computeService.Compute(model);

            if (resultAlgoritmRoutes.FailedToFindRoute) return Redirect(Url.Action("RouteNotFound", "Home"));

            return View(resultAlgoritmRoutes);
        }

        public IActionResult RouteNotFound()
        {

            return View(true);
        }

        private List<RouteViewModel> GetRoutes(Db.Model.TravelMode m)
        {
            return _dbContext.Routes.Select(d => new RouteViewModel()
            {
                Id = d.Id,
                Destination = GetPlaceNameById(d.DestinationId),
                Origin = GetPlaceNameById(d.OriginId),
                DurationText = GetlenghtText(d.MeasuresLenght.SingleOrDefault(b => b.TravelMode == m)?.Duration ?? 0, "Duration"),
                DistanceText = GetlenghtText(d.MeasuresLenght.SingleOrDefault(b => b.TravelMode == m)?.Distance ?? 0, "Distance")
            }).ToList();


            string GetlenghtText(int length, string routeWeightValue)
            {
                if (routeWeightValue == WeightMode.Duration.ToString())
                {
                    TimeSpan routeTime = TimeSpan.FromMinutes(length / 60);
                    if(length / 60 / 60 / 24 >= 1)
                    {
                        return $"{routeTime.Days} d {routeTime.Hours} h {routeTime.Minutes} min";
                    }
                    else if (length / 60 / 60 >= 1)
                    {
                        return $"{routeTime.Hours} h {routeTime.Minutes} min";
                    }
                    else
                    {
                        return $"{routeTime.Minutes} min";
                    }
                }
                else
                {
                    double kmlenght = Math.Round(length / 1000d, 1);
                    return $"{kmlenght:0.0} km".Replace(".", ",");
                }
            }
        }

        private string GetPlaceNameById(int destinationId)
        {
            return _dbContext.Places.SingleOrDefault(d => d.Id == destinationId)?.NickName ?? "";
        }

        [ResponseCache(Duration = 0, Location = ResponseCacheLocation.None, NoStore = true)]
        public IActionResult Error()
        {
            return View(new ErrorViewModel { RequestId = Activity.Current?.Id ?? HttpContext.TraceIdentifier });
        }
    }
}
