﻿using Optiroute.Db;
using Optiroute.Db.Model;
using Optiroute.TSPAlgorithm;
using Optiroute.Web.Models.Home;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using System.Diagnostics;

namespace Optiroute.Web.Services
{
    public interface IComputeAlgorithmRoutesService
    {
        ResultRoutesViewModel Compute(IndexViewModel model);
    }
    public class ComputeAlgorithmRoutesService : IComputeAlgorithmRoutesService
    {
        private readonly ITSPComputeAlgorithmProviderService _tspComputeRoutes;
        private readonly ITSPOptimaliseAlgorithmProviderService _tspOpimaseRoutes;
        private readonly DbContext _dbContext;
        private readonly ICodeBookService _codebookservice;

        public ComputeAlgorithmRoutesService(ITSPComputeAlgorithmProviderService tspComputeRoutes, ITSPOptimaliseAlgorithmProviderService tspOpimaseRoutes, DbContext dbContext, ICodeBookService codebookservice)
        {
            this._tspComputeRoutes = tspComputeRoutes;
            this._tspOpimaseRoutes = tspOpimaseRoutes;
            this._dbContext = dbContext;
            this._codebookservice = codebookservice;
        }
        private int GetLength(IndexViewModel model, Db.Model.Route route)
        {
            var r = route.MeasuresLenght.SingleOrDefault(d => d.TravelMode.ToString() == model.TravelModeValue);
            return model.RouteWeightValue == WeightMode.Duration.ToString() ? r.Duration : r.Distance;
        }

        public ResultRoutesViewModel Compute(IndexViewModel model)
        {
            MatrixRoute matrix = GetMatrix(model);
            ResultRoutesViewModel resultRoutes = new ResultRoutesViewModel() {
                LenghtUnits = model.RouteWeightValue == WeightMode.Duration.ToString() ? "min" : "m",
                WeightIcon = _codebookservice.GetWeightModes().SingleOrDefault(d => d.Mode.ToString() == model.RouteWeightValue).WeightIcon,
                TravelModeIcon = _codebookservice.GetTravelModes().SingleOrDefault(d => d.Travel.ToString() == model.TravelModeValue).TravelIcon,
                FailedToFindRoute = RecognizeFailToFindRoute(matrix)
            };

            if (resultRoutes.FailedToFindRoute) return resultRoutes;
            foreach (var a in _tspComputeRoutes.GetAllComputeAlgorithms())
            {
                RouteTSPResult tspRoute = a.ComputeTSPRoute(matrix);
                resultRoutes.ResultRoutes.Add(NewMethod(model,tspRoute, a.Name, a.ShortName));
                //Doplňování pro optimalizační algorimy
                foreach (var k in _tspOpimaseRoutes.GetAllOptimalizationAlgorithms())
                {
                    RouteTSPResult optiRoute = k.OptimaliseTSPPRoute(matrix, tspRoute);
                    resultRoutes.ResultRoutes.Add(NewMethod(model, optiRoute, $"{k.Name}: {a.Name}", $"{k.ShortName}: {a.ShortName}"));
                }
            }

            int lenghtRoute = resultRoutes.ResultRoutes.Min(d => d.LenghtRoute);
            foreach (var a in resultRoutes.ResultRoutes.Where(b => b.LenghtRoute == lenghtRoute))
            {
                a.ShortestRoute = true; 
            }

            long clockedTime = resultRoutes.ResultRoutes.Min(d => d.ClockedTime);
            foreach (var a in resultRoutes.ResultRoutes.Where(b => b.ClockedTime == clockedTime))
            {
                a.FastestAlgorithm = true;
            }
            return resultRoutes;
        }

        private ResultRoute NewMethod(IndexViewModel model, RouteTSPResult tspRoute, string name, string shortName)
        {
            List<Models.Home.ResultRoute.Route> routesWithTime = new List<Models.Home.ResultRoute.Route>();
            for (int i = 0; i < tspRoute.AddressisRoute.Count(); i++)
            {
                if (tspRoute.AddressisRoute[i] == tspRoute.AddressisRoute[0] && i != 0) continue;
                var dbRoute = _dbContext.Routes.SingleOrDefault(d => (d.OriginId == tspRoute.AddressisRoute[i] && d.DestinationId == tspRoute.AddressisRoute[i + 1 < tspRoute.AddressisRoute.Count() ? i + 1 : 0]) || (d.DestinationId == tspRoute.AddressisRoute[i] && d.OriginId == tspRoute.AddressisRoute[i + 1 < tspRoute.AddressisRoute.Count() ? i + 1 : 0]));
                var dbRouteMeasure = dbRoute?.MeasuresLenght.SingleOrDefault(d => d.TravelMode.ToString() == model.TravelModeValue);
                if ((dbRoute?.OriginId == _dbContext.Places.SingleOrDefault(d => d.Default).Id && dbRoute?.DestinationId == _dbContext.Places.SingleOrDefault(d => d.EndCity)?.Id) || (dbRoute?.OriginId == _dbContext.Places.SingleOrDefault(d => d.EndCity)?.Id && dbRoute?.DestinationId == _dbContext.Places.SingleOrDefault(d => d.Default).Id))
                {
                    dbRouteMeasure.Distance = 0;
                    dbRouteMeasure.Duration = 0;
                }
                routesWithTime.Add(new ResultRoute.Route()
                {
                    OriginName = _dbContext.Places.Single(d => d.Id == tspRoute.AddressisRoute[i]).NickName,
                    OriginEndCity = _dbContext.Places.Single(d => d.Id == tspRoute.AddressisRoute[i]).EndCity,
                    DestinationName = _dbContext.Places.Single(d => d.Id == tspRoute.AddressisRoute[i + 1 < tspRoute.AddressisRoute.Count() ? i + 1 : 0]).NickName,
                    LenghtText = GetlenghtText(model.RouteWeightValue == WeightMode.Duration.ToString() ? dbRouteMeasure.Duration : dbRouteMeasure.Distance, model.RouteWeightValue)
                });
            }

           return new ResultRoute()
            {
                Name = name,
                Shortname = shortName,
                LenghtRoute = model.RouteWeightValue == WeightMode.Duration.ToString() ? tspRoute.Length/60 : tspRoute.Length,
                ClockedTime = tspRoute.TimeCalculation,
                LenghtRouteText = GetlenghtText(tspRoute.Length, model.RouteWeightValue),
                AlgorithmRoute = tspRoute.AddressisRoute.Select(d => _dbContext.Places.Single(b => b.Id == d).NickName).ToList(),
                RoutesWithTime = routesWithTime
            };
        }

        private bool RecognizeFailToFindRoute(MatrixRoute matrix)
        {
            bool failed = false;
            for (int i = 0; i < matrix.placeIDs.Count(); i++)
            {
                for (int j = 0; j < matrix.placeIDs.Count(); j++)
                {
                    if (matrix.matrixRoute[i, j] == -1) failed = true;
                }
            }
            return failed;
        }

        string GetlenghtText(int length, string routeWeightValue)
        {
            if (routeWeightValue == WeightMode.Duration.ToString())
            {
                TimeSpan routeTime = TimeSpan.FromMinutes(length / 60);
                if (length / 60 / 60 / 24 >= 1)
                {
                    return $"{routeTime.Days} d {routeTime.Hours} h {routeTime.Minutes} min";
                }
                else if (length / 60 / 60 >= 1)
                {
                    return $"{routeTime.Hours} h {routeTime.Minutes} min";
                }
                else
                {
                    return $"{routeTime.Minutes} min";
                }
            }
            else
            {
                double kmlenght = Math.Round(length / 1000d, 1);
                return $"{kmlenght:0.0} km".Replace(".", ",");
            }
        }

        private MatrixRoute GetMatrix(IndexViewModel model)
        {
            List<int> checkedAddressIDs = model.Addresses.Where(d => d.Selected == true).Select(d => d.Id).ToList();
            checkedAddressIDs.Add(_dbContext.Places.Single(d => d.Default).Id);

            List<Optiroute.Db.Model.Route> requestedRoutes = _dbContext.Routes.Where(d => checkedAddressIDs.Contains(d.OriginId) && checkedAddressIDs.Contains(d.DestinationId)).ToList();

            if (model.Addresses.Where(d => d.EndCity).ToList().Count > 1)
            {
                throw new ApplicationException("Není možné mít vybrané více konečných poboček");
            }

            InputAlgorithmRoute inputAlgorithmRoute = new InputAlgorithmRoute()
            {
                InputRoutes = requestedRoutes.Select(d => new TSPAlgorithm.Route()
                {
                    OriginID = d.OriginId,
                    DestinationId = d.DestinationId,
                    Length = GetLength(model, d)

                }).ToList(),
                DefaultPlaceID = _dbContext.Places.Single(d => d.Default && checkedAddressIDs.Contains(d.Id)).Id,
                EndPlaceID = _dbContext.Places.SingleOrDefault(d => d.EndCity && checkedAddressIDs.Contains(d.Id))?.Id
            };

            return new MatrixRoute(inputAlgorithmRoute);
        }
    }
}
